(* file: queens.ml
 * author: Bob Muller
 *
 * CS1101 Computer Science I
 * 
 * This program solves the 8-queens problem. It uses Asai's Universe library.
 *
 * Compile and run:
 *
 * > make
 *> ./queens
 *)
open World
open Color
open Image
open Cs1101

let imageSide = 100.               (* determined by size of crown.png *)
let boardSize = 8. *. imageSide
			
let emptyBoardImage = Image.rectangle boardSize boardSize Color.white

let whiteSquare = Image.rectangle imageSide imageSide Color.darkGray
let graySquare  = Image.rectangle imageSide imageSide Color.gray20

type piece_t  = Queen | Empty

type board_t = (piece_t array) array
type stack_t = (int * int) Stack.t
type world_t = {board:board_t; stack:stack_t; row:int; col:int; solved:bool}

let next (row, col) =         (* starts at (0, 0); returns *)
  if col < 7 then             (* (8, 0) when all done.     *)
    (row, col + 1)
  else
    (row + 1, 0)
    
let attacksFromAbove (row1, col1) (row2, col2) board =
  let hasAQueen  = board.(row1).(col1) = Queen in
  let sameColumn = col1 = col2 in
  let onDiagonal = (abs (row1 - row2)) = (abs (col1 - col2))
  in
  hasAQueen && (sameColumn || onDiagonal)

let attackedFromAbove (row, col) board =
  let rec repeat (row', col') =
    match row' = row with
    | true  -> false
    | false -> (match attacksFromAbove (row', col') (row, col) board with
		| true  -> true
		| false -> repeat (next(row', col')))
  in
  repeat (0, 0)

let move ({board; stack; row; col} as world) =
  let rec solve (row, col) =
    match (row = 8) with
    | true  -> World({ world with solved = true})
    | false -> (match (attackedFromAbove (row, col) board) with
		| true  -> (match col < 7 with
			    | true  -> solve (row, col + 1)
			    | false -> backTrack world)
		| false -> board.(row).(col) <- Queen;
			   Stack.push (row, col) stack;
			   World.World({world with row=row+1; col=0}))

  and backTrack world =
    match Stack.is_empty world.stack with
    | true  -> failwith "queens: there is no solution.\n"
    | false -> let (row', col') = Stack.pop world.stack
	       in
	       board.(row').(col') <- Empty;
	       match col' = 7 with
	       | true  -> backTrack world               (* keep popping ... *)
	       | false -> solve (next (row', col'))
  in
  solve (row, col)
	
let draw crown {board; stack} =
  let getImage piece row col =
    match piece with
    | Empty -> (if ((row + col) mod 2) = 0 then whiteSquare else graySquare)
    | Queen -> crown in
  let rec makeBoardImage (row, col) image =
    match row = 8 with
    | true  -> image
    | false -> let img = getImage board.(row).(col) row col in
	       let (x, y) = ((i2F col) *. imageSide, ((i2F row) *. imageSide)) in 
	       let newImage = Image.place_image img (x, y) image
	       in
	       makeBoardImage (next (row, col)) newImage
  in
  makeBoardImage (0, 0) emptyBoardImage

let finished world = world.solved

let queens () =
  let crown = Image.read_image "crown.png" in
  let board = Array.make_matrix 8 8 Empty in
  let initialWorld = {board=board; stack=Stack.create (); row=0; col=0; solved=false}
  in
  World.big_bang initialWorld
		 ~name:"Eight Queens!"
		 ~width:(f2I boardSize)
		 ~height:(f2I boardSize)
		 ~on_tick:move  
		 ~rate:0.5
		 ~to_draw:(draw crown) 
		 ~stop_when:finished
		 ~to_draw_last:(draw crown)

let () = queens ()
