(* file: shannonMarkov.ml
 * author: Bob Muller, adapted from a program by David Martin
 * date: November 20, 2014
 * 
 * This program implements the N-gram text generation system
 * proposed by Shannon in "A Mathematical Theory of Communication" 
 * (1948).
 * 
 * Run this app using:
 * 
 * ocaml -I +labltk str.cma unix.cma labltk.cma jpflib.cma
 *
 * ocamlc -I +labltk str.cma unix.cma labltk.cma jpflib.cma shannonMarkov.ml
 *)
open Tk;;

(* Code for implementing a map relating keys (i.e., strings) to
 * lists of chars. See the API for Map in OCaml's Standard Library.
 *)
module Map = Map.Make(String);;

let makeMap () = Map.empty;;

let mapToString map =
  let charListToString v = 
    "[" ^ (String.concat "; " (List.map Char.escaped v)) ^ "]" in
  let kvStr key cl rest = ("\n" ^ key ^ "=" ^ (charListToString cl) ^ rest) in
    "{" ^ (Map.fold kvStr map "}");;

(* val getText : textWidget -> string
 *
 * Retrieves input text as a string from the GUI.
 *)
let getText textField : string =
  Text.get ~start:(`Linechar (0,0),[]) ~stop:(`End,[]) textField;;

(* val putText : string -> textWidget -> unit
 *
 * (putText text outputText) puts text into the provided text widget.
 *)
let putText text outputText =
  Text.insert ~index:(`End,[]) ~text:text outputText;;

let clearOutput textField =
  Text.delete ~start:(`Linechar (0,0),[]) ~stop:(`End,[]) textField;;

(* val getDegree : entryWidget -> int
 *
 * Retrieves the degree from the entry field of the GUI.
 *)
let getDegree degreeEntry = int_of_string (Entry.get degreeEntry);;

(* val extend : string -> char -> int 
 *
 * (extend string char degree) returns answer=string+c ensuring that
 * answer is no longer than degree by possibly dropping the leftmost
 * letter.
 *)
let extend string char degree =
  let charS : string = String.make 1 char in
  let answer = String.concat "" [string; charS] in
    match (String.length answer) <= degree with
	true  -> answer
      | false -> String.sub answer 1 degree;;

(* val addChar : char -> string -> map -> map
 *
 * (addChar chr key map) map[key] = [c1; ...; ck], addChar adds
 * chr to the list returning the new map.
*)
let addChar chr key map =
  match Map.mem key map with
      true ->
	let chrs = (Map.find key map) @ [chr] 
	in
	  Map.add key chrs (Map.remove key map)
    | false -> map;;

(* val sentinal : char  --- the special value used for the end of 
 * the input.
 *)
let sentinal = Char.chr 0;;

let buildModel degree inputText =
  let text : string = getText inputText in
  let rec loop prefix i map =
    match i = (String.length text) with
    | true  -> Map.add prefix [sentinal] map
    | false ->
       let nextLetter : char = text.[i] in
       let newMap = 
	 match Map.mem prefix map with
	 | true  -> addChar nextLetter prefix map
	 | false -> Map.add prefix [nextLetter] map
       in
       loop (extend prefix nextLetter degree) (i + 1) newMap in
  let emptyMap = makeMap()
  in
  loop "" 0 emptyMap;;

(* val pickALetter : string -> (char list) map -> char
 *
 * (pickALetter key map), let map[key] = [c1; ...; ck], pickALetter
 * returns a randomly chosen character from the list.
*)
let pickALetter key model =
  let cl = Map.find key model in
  let k = Random.int (List.length cl) in
    List.nth cl k;;

let generateOutput model degree outputText = 
  clearOutput outputText;
  let rec loop prefix letter =
    match letter = sentinal with
	true -> ()
      | false -> 
	  putText (Char.escaped letter) outputText;
	  let newPrefix = extend prefix letter degree in
	  let nextLetter = pickALetter newPrefix model 
	  in
	    loop newPrefix nextLetter in
  let initialPrefix = "" in
  let firstLetter = pickALetter initialPrefix model
  in
    loop initialPrefix firstLetter;;

  
let shannonMarkov inputText degree outputText =
  let model = buildModel degree inputText in
    print_endline (mapToString model);
    generateOutput model degree outputText;;

  
(* val writeFile : string -> textWidget -> unit
 *
 * The call (writeFile fileName outputText) opens file name
 * and write the contents of the outputText field to that file.
 *)
let writeFile ~file:fileName outputText =
  let ouch = open_out fileName in
  let outString = getText outputText in
    output_string ouch outString;
    flush ouch;
    close_out ouch;
    ();;

(* val buildGUIAndGo : unit -> unit
 *
 * The main entry point of an implementation of Shannon's n-gram
 * system. The application uses OCaml's Labltk module, plus a few
 * others. See the instructions at the top of this file for running
 * the app.
 *)
let buildGUIAndGo () =
  let top = openTk() in
  let _ = Wm.title_set top "Shannon's Markov Model of Text" in
  let leftFrame   = Frame.create top in 
  let middleFrame = Frame.create top in
  let rightFrame  = Frame.create top in
  let bgColor = `Color "#CACACA" in       (* light grey *)

  (* The lefthand side is for input text. 
  *)
  let inputText = Text.create ~background:bgColor leftFrame in
  let fileselect ~dir:dir = 
    Fileselect.f ~title:"File Selector" ~multi:false ~sync:false
      ~filter:(Filename.concat dir "*.txt")
      ~file:"."
      ~action:(fun files ->
		 match files with
		     [] -> ()
		   | [fileName] -> 
		       prerr_endline ("selected ``" ^ fileName ^ "''");
		       (* Open the selected file
		       *)
		       let inputFile = open_in fileName in
			 (try
			    while true do
			      let s = input_line inputFile in
				Text.insert ~index:(`End,[]) ~text:s inputText
			    done
			  with
			      End_of_file -> close_in inputFile)
		 | _ -> raise (Failure "multi switched off")) 

  in
  let openInputFile ignored = fileselect ~dir:(Sys.getcwd ()) in
  let openButton = Button.create ~text:"Open" ~command:openInputFile leftFrame in

  (* The middle frame has an entry and a button for running the app.
  *)
  let degreeVar = Textvariable.create () in Textvariable.set degreeVar "4";
  let degreeEntry = Entry.create ~width:10 ~textvariable:degreeVar middleFrame in 

  let outputText = Text.create ~background:bgColor rightFrame in
  let generate ignored = 
    let degree = getDegree degreeEntry in
      shannonMarkov inputText degree outputText in
  let generateButton = Button.create ~text:">>>" ~command:generate middleFrame in
  let writeToFile ignored = writeFile ~file:"out.txt" outputText in
  let saveButton = Button.create ~text:"Save" ~command:writeToFile rightFrame 
  in
    pack ~side:`Left [leftFrame; middleFrame; rightFrame];
    pack [openButton];
    pack [inputText];
    pack [coe degreeEntry];
    pack [generateButton];
    pack [rightFrame]; 
    pack [saveButton];
    pack [outputText];
    Printexc.print mainLoop();;

let _ = buildGUIAndGo();;
