(* file: svm.ml
 * author: Bob Muller
 * date: October 26, 2014, adapted from svm.py
 * revised: March 2015, added Mov, Mul and Div instructions.
 *
 *
 * This program implements the Simple Virtual Machine (SVM).
 *)

(* A type for the register file. The SVM has 4 general purpose registers.
*)
type registers = {r0 : int; r1 : int; r2 : int; r3 : int};;

let string_of_regs {r0; r1; r2; r3} =
  let r0s = string_of_int r0 in
  let r1s = string_of_int r1 in
  let r2s = string_of_int r2 in
  let r3s = string_of_int r3 in
    "{r0=" ^ r0s ^ "; r1=" ^ r1s ^ "; r2=" ^ r2s ^ "; r3=" ^ r3s ^ "}";;

(* Names for registers.
*)
type reg = R0 | R1 | R2 | R3;;

let string_of_reg r = 
  match r with R0 -> "R0" | R1 -> "R1" | R2 -> "R2" | R3 -> "R3";;

let registerGet register {r0; r1; r2; r3} =
  match register with R0 -> r0 | R1 -> r1 | R2 -> r2 | R3 -> r3;;

let registerPut value dest regs =
  match dest with
      R0 -> { regs with r0=value}
    | R1 -> { regs with r1=value}
    | R2 -> { regs with r2=value}
    | R3 -> { regs with r3=value};;

(* These opnd types to satisfy OCaml's requirement for type declarations
 * for record types. 
 *)
type opnd0 = {rd : reg; rs : reg};;
type opnd1 = {rd : reg; addr : int};;
type opnd2 = {rs : reg; addr : int};;
type opnd3 = {rd : reg; rs : reg; rt : reg};;
type opnd4 = {rs : reg; rt : reg};;

type instruction = 
    Lod of opnd1  (* {rd : reg; addr : int} *)
  | Sto of opnd2  (* {rs : reg; addr : int} *)
  | Mov of opnd0  (* {rd : reg; rs : reg} *)
  | Add of opnd3  (* {rd : reg; rs : reg; rt : reg} *)
  | Sub of opnd3  (* ditto *)
  | Mul of opnd3  (* ditto *)
  | Div of opnd3  (* ditto *)
  | Cmp of opnd4  (* {rs : reg; rt : reg } *)
  | Cpz of reg 
  | Blt of int 
  | Beq of int
  | Bgt of int
  | Jmp of int
  | Hlt;;

let string_of_instruction i =
  let si = string_of_int in
  let sr = string_of_reg in
  match i with
    Lod {rd; addr} -> "Lod\t" ^ (sr rd) ^ ", " ^ (si addr)
  | Sto {rs; addr} -> "Sto\t" ^ (sr rs) ^ ", " ^ (si addr)
  | Mov {rd; rs} -> "Mov\t" ^ (sr rd) ^ ", " ^ (sr rs)
  | Add {rd; rs; rt} -> "Add\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Sub {rd; rs; rt} -> "Sub\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Mul {rd; rs; rt} -> "Mul\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Div {rd; rs; rt} -> "Div\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Cmp {rs; rt} -> "Cmp\t" ^ (sr rs) ^ ", " ^ (sr rt)
  | Cpz reg -> "Cpz\t" ^ (sr reg)
  | Blt disp -> "Blt\t" ^ (si disp)
  | Beq disp -> "Beq\t" ^ (si disp)
  | Bgt disp -> "Bgt\t" ^ (si disp)
  | Jmp disp -> "Jmp\t" ^ (si disp)
  | Hlt -> "Hlt";;
      
(* Object files have data segments and text segments. Representing 
 * segments as lists here. Pretty inefficient!
 *)
type datasegment = int list;;
type textsegment = instruction list;;

type image = {data : datasegment; text : textsegment};;

let split segment n =
  let rec loop n xs ys =
    match (n, xs) with
	(0, _) -> (List.rev ys, xs)
      | (n, z::zs) -> loop (n - 1) zs (z::ys) 
  in loop n segment [];;

let ramGet n segment = List.nth segment n;;
let ramPut v n segment =                     (* Wildly inefficient! *)
  let (left, right) = split segment n in
    left @ (v :: right);;

(* A sample SVM program. Let M = data1[0], N = data1[1].
 * Then program1 computes R0 = M * N.
 *)
let data1 = [4; 5; 0; 1];;
let text1 : instruction list = 
  [ 
    Lod {rd=R0; addr=2};
    Lod {rd=R3; addr=3};
    Lod {rd=R1; addr=0};
    Lod {rd=R2; addr=1};
    Cpz R2;
    Beq 3;
    Add {rd=R0; rs=R0; rt=R1};
    Sub {rd=R2; rs=R2; rt=R3};
    Jmp (-5);
    Hlt
  ];;
let program1 : image = {data=data1; text=text1};;

(* Program pgm is typically executed on svm with the call: (svm pgm 0)
 *)
let debug = true;;
let step = false;;

let printState pc psw registers =
  let pcs = string_of_int pc in
  let psws = string_of_int psw in
  let regstr = string_of_regs registers in
    begin
      print_endline ("\ndbg: (pc, psw) = (" ^ pcs ^ ", " ^ psws ^ ")");
      print_endline ("dbg: regs = " ^ regstr)
    end;;

(* An implementation of the Simple Virtual Machine. A call (svm ram pc)
 * executes the SVM program starting with instruction ram[pc].
 *)
let svm ram pc =

  (* The CPU Instruction cycle.
  *)
  let rec cycle pc psw registers ({data; text} as ram) =

    (* Fetch the next instruction from RAM.
     *)
    let instruction = ramGet pc text in
    let newPC = pc + 1 in
    let _ = match (debug, step) with
      | (true, true) ->
	  begin
	    printState pc psw registers;
	    read_line();
	    ()
	  end
      | (true, false) -> printState pc psw registers
      | _ -> ()
    in
      (* Now dispatch on the instruction.
      *)
      match instruction with

	  Lod {rd; addr} ->
	    let value = ramGet addr data in
	    let newRegisters = registerPut value rd registers in
              cycle newPC psw newRegisters ram

	| Sto {rs; addr} ->
	    let value = registerGet rs registers in
	    let newData = ramPut value addr data in
              cycle newPC psw registers { ram with data = newData }

	| Mov {rd; rs} ->
	    let v = registerGet rs registers in
	    let newRegisters = registerPut v rd registers in
	      cycle newPC psw newRegisters ram

	| Add {rd; rs; rt} -> 
	    let v1 = registerGet rs registers in
	    let v2 = registerGet rt registers in
	    let newRegisters = registerPut (v1 + v2) rd registers in
              cycle newPC psw newRegisters ram

	| Sub {rd; rs; rt} ->
	    let v1 = registerGet rs registers in
	    let v2 = registerGet rt registers in
	    let newRegisters = registerPut (v1 - v2) rd registers in
              cycle newPC psw newRegisters ram

	| Mul {rd; rs; rt} ->
	    let v1 = registerGet rs registers in
	    let v2 = registerGet rt registers in
	    let newRegisters = registerPut (v1 * v2) rd registers in
              cycle newPC psw newRegisters ram

	| Div {rd; rs; rt} ->
	    let v1 = registerGet rs registers in
	    let v2 = registerGet rt registers in
	    let newRegisters = registerPut (v1 / v2) rd registers in
              cycle newPC psw newRegisters ram

	| Cmp {rs; rt} ->
            let v1 = registerGet rs registers in
	    let v2 = registerGet rt registers in
	    let value = v1 - v2 in
	      cycle newPC value registers ram

	| Cpz r ->
	    let value = registerGet r registers in
	      cycle newPC value registers ram

	| Blt displacement ->
	    (match psw < 0 with
		 true  -> cycle (newPC + displacement) psw registers ram
	       | false -> cycle newPC psw registers ram)

	| Beq displacement ->
	    (match psw = 0 with
		 true  -> cycle (newPC + displacement) psw registers ram
	       | false -> cycle newPC psw registers ram)

	| Bgt displacement ->
	    (match psw > 0 with
		 true  -> cycle (newPC + displacement) psw registers ram
	       | false -> cycle newPC psw registers ram)

	| Jmp displacement ->
            cycle (newPC + displacement) psw registers ram

	| Hlt -> 
	    begin
              print_endline "\nSVM Halt.";
	      printState pc psw registers
	    end 
  in
  let initialPSW = 0 in
  let initialRegisters = {r0=0; r1=0; r2=0; r3=0} in
    cycle pc initialPSW initialRegisters ram;;
